
open Sasacore
(* open SasArg *)
open RdbgPlugin

(* currently useless as we only read bool (for now) *)
let rec (_to_sasa_value : Data.v -> Algo.value) = function 
  | Data.I i -> Algo.I i
  | Data.F f -> Algo.F f
  | Data.B b  -> Algo.B b
  | Data.E(_id, i) -> Algo.E i
  | Data.A a -> Algo.A (Array.map _to_sasa_value a)
  | Data.S _s -> assert false
  | Data.U -> assert false 

let rec (of_sasa_value : Algo.value -> Data.v) = function 
  | Algo.I i -> Data.I i
  | Algo.F f -> Data.F f
  | Algo.B b -> Data.B b
  | Algo.E i -> Data.E(string_of_int i,i)
  | Algo.S _s -> assert false
  | Algo.N i -> Data.I i
  | Algo.A a -> Data.A (Array.map of_sasa_value a) 
  
let (get_action_value : (string * Data.v) list -> string -> string -> bool) =
  fun sl pid a ->
    let vn = Printf.sprintf "%s_%s" pid a in
    match List.assoc_opt vn sl with
    | Some (Data.B b) -> b
    | None 
    | Some _ ->
      Printf.printf "Error: can not find %s in [%s]\n" vn
        (String.concat "," (fst (List.split sl)));
      flush stdout;
      assert false

(*
let (to_sasa_env : (string * Data.v) list -> Env.t) =
  fun sl ->
    (* inputs are actions of the form "pid_ActionName" *)
    let split_var str =
      match Str.split (Str.regexp "_") str with
      | pid::vn  -> pid, String.concat "_" vn
      | [] -> assert false
      | [_] -> assert false
      | _ -> assert false
    in
    let _sl = List.map (fun (var,v) -> split_var var, to_sasa_value v) sl in
    assert false
*)
        
open Process
let (from_sasa_env : Sasa.layout -> Env.t -> RdbgPlugin.sl) =
  fun p_nl_l e ->
    let sll =
      List.map
        (fun (p,_nl) ->
           List.map (fun (vn,_) ->
               Printf.sprintf "%s_%s" p.pid vn, Env.get e p.pid vn)
             p.variables)
        p_nl_l
    in
    let sl = List.flatten sll in
    List.map (fun (vn,value) -> vn, of_sasa_value value) sl

let (get_sl_out: bool -> Process.t list -> bool list list -> RdbgPlugin.sl) =
  fun enab pl ll ->
    let str = if enab then "Enab_" else "" in
    List.flatten (
      List.map2 (fun p enab_l ->
          List.map2 (fun a enab ->
              Printf.sprintf "%s%s_%s" str p.pid a, Data.B enab)
            p.actions enab_l)
        pl ll
    )

(* let (bool: Algo.value -> bool) = function B b -> b | _ -> assert false *)

let (make_do: string array -> SasArg.t ->
     (Process.t * Algo.neighbor list) list -> Env.t -> RdbgPlugin.t) =
  fun argv args pl_n e -> 
    let pl = fst (List.split pl_n) in
    let prog_id = Printf.sprintf "%s (with sasa Version %s)"
        (String.concat " " (Array.to_list argv)) SasaVersion.str
    in
    let vntl_i =
      List.map (fun (vn,vt) -> vn, Data.type_of_string vt)
        (Sasa.get_inputs_rif_decl args pl)
    in
    let vntl_o =
      List.map (fun (vn,vt) -> vn, Data.type_of_string vt)
        (Sasa.get_outputs_rif_decl args pl)
    in
    let pre_enable_processes_opt = ref None in
    let sasa_env = ref e in
    let reset () =
      pre_enable_processes_opt := None;
      sasa_env := e
    in
    (*  Do the same job as SasaMain.simustep *)
    let (step: RdbgPlugin.sl -> RdbgPlugin.sl) =
      fun sl_in ->
        let e = !sasa_env in
        match !pre_enable_processes_opt with
        | None  -> (* the first step *)
          (* 1: Get enable processes *)
          let pnall, enab_ll = Sasa.get_enable_processes pl_n e in
          let sasa_nenv = from_sasa_env pl_n e in 
          pre_enable_processes_opt := Some(pnall, enab_ll);
          sasa_nenv @ (get_sl_out true pl enab_ll) 
        | Some (pre_pnall, pre_enab_ll) ->
          (* 2: read the actions from the outside process, i.e., from sl_in *)
          let _, pnal = Demon.f args.dummy_input
              (args.verbose > 1) args.demon pl pre_pnall pre_enab_ll
              (get_action_value sl_in)
          in
          (* 3: Do the steps *)
          let ne = Sasa.do_step pnal e in
          let sasa_nenv = from_sasa_env pl_n ne in
          (* 1': Get enable processes *)
          let pnall, enab_ll = Sasa.get_enable_processes pl_n ne in
          pre_enable_processes_opt := Some(pnall, enab_ll);
          sasa_env := ne;
          sasa_nenv @ (get_sl_out true pl enab_ll) 
    in
    let (step_internal_demon: RdbgPlugin.sl -> RdbgPlugin.sl) =
      fun sl_in ->
        (* in this mode, sasa does not play first *)
        let e = !sasa_env in
        (* 1: Get enable processes *)
        let pnall, enab_ll = Sasa.get_enable_processes pl_n e in
        (* 2: read the actions from the outside process, i.e., from sl_in *)
        let activate_val, pnal = Demon.f args.dummy_input
            (args.verbose > 1) args.demon pl pnall enab_ll
            (get_action_value sl_in)
        in
        (* 3: Do the steps *)
        let ne = Sasa.do_step pnal e in
        sasa_env := ne;
        (from_sasa_env pl_n e) @ (get_sl_out true pl enab_ll) @
        (get_sl_out false pl activate_val) 
    in
    let step = if args.demon = Demon.Custom then step else step_internal_demon in
    let ss_table = Hashtbl.create 10 in
    let step_dbg sl_in ctx cont = cont (step sl_in) ctx in
(*
   this event is useless actually; the same information is available at Rtop
      { ctx with
        Event.kind = Event.Exit;
        Event.lang = "sasa";
        Event.sinfo = None;
        Event.name = prog_id;
        Event.inputs = [] ;  
        Event.outputs = [];
        Event.locals = [];
        Event.next = (
          fun () ->  
            let ctx = { ctx with Event.nb = ctx.Event.nb+1 } in
            cont (step sl_in) ctx );
      }
    in
*) 
    let (mems_in  : Data.subst list) = [] in
    let (mems_out : Data.subst list) = [] in
    {
      id = prog_id;
      inputs = vntl_i;
      outputs= vntl_o;
      reset=(fun () -> reset());
      kill=(fun _ -> flush stdout; flush stderr);
      init_inputs=mems_in;
      init_outputs=mems_out;
      step=step;     
      step_dbg = step_dbg;
      save_state = (fun i ->
          let prgs = Random.get_state () in
          Hashtbl.replace ss_table i
            (prgs, !sasa_env, !pre_enable_processes_opt)
        );
      restore_state = (fun i ->
          match Hashtbl.find_opt ss_table i with
          | Some (prgs, e, pepo) ->
            Random.set_state prgs;
            sasa_env := e; pre_enable_processes_opt := pepo
          | None ->
            Printf.eprintf "Cannot restore state %i from sasa\n" i;
            flush stderr 
        );
    }


let (make: string array -> RdbgPlugin.t) =
  fun argv ->
    try
      let args, pl_n, e = Sasa.make false argv in
      make_do argv args pl_n e
    with
    | Dynlink.Error e ->
      Printf.printf "Error: %s\n" (Dynlink.error_message e);
      flush stdout;
      exit 2
    | Failure msg ->
      Printf.printf "Error: %s\n" msg;
      flush stdout;
      exit 2

    | Assert_failure (file, line, col)  -> 
      prerr_string (
        "\nError: oops, sasa internal error\n\tFile \""^ file ^
        "\", line " ^ (string_of_int line) ^ ", column " ^
        (string_of_int col) ^ "\n") ; 
      flush stderr;
      exit  2
